-------------------------------------------------------------------------------
--Aggregator version 5.
-------------------------------------------------------------------------------
--One-time pad for hash table mixer
--Changed cypher into pad... speeds up hash calculation further
--Flat bucket approach (no bucketlets)
--Ordinary sequence (with bubble sort) for ranking section
--Stores frequency counts as negative, to increase speed of ranker
--Uses a single sequence to hold freq, raw, word
--Short-circuits comparing the whole word if the raw hashes are different
-------------------------------------------------------------------------------

without type_check
without warning

--<AGGREGATOR>=================================================================
--The aggregator takes the words from the parser, and generates hash tables
--containing frequency use.
--Aggregate statistics are calculated here, too.
--Optimised where feasible.
-------------------------------------------------------------------------------
--Internal constants
global constant TABLE_SIZE = 4097       --Some random number that seems to work well
global constant TABLE_WIDTH = 500       --The bucket capacity
--(Total space in table is TABLE_SIZE * TABLE_WIDTH)

--Subscript identifiers: frequency, raw_hash, and word.
global constant FREQ = 1, 
                RAW  = 2,
                WORD = 3
-------------------------------------------------------------------------------
--Internal variables (global for debugging)
global sequence hash_table --A hash table holding many buckets. Each bucket can contain up to TABLE_WIDTH buckets
global sequence table_width --A table holding the collision count for each bucket

--Statistics
global integer word_total           --Total number of words
global integer unique_total         --Total number of unique words
global sequence totals_by_length    --Total number for each length of word

--Internal temporary variables
sequence word               --New words are stored in here
atom raw_hash               --The unmodulii'd hash value
atom hash                   --Bucket number the word belongs in
integer is_unique           --Whether the word has been found yet
integer index_in_bucket     --The location of the word in a bucket, if it exists
integer word_length         --Length of the word, 0-20
-------------------------------------------------------------------------------
--Initialise tables
hash_table = repeat( repeat(0,TABLE_WIDTH),TABLE_SIZE ) --The initial value is not important
table_width = repeat( 0, TABLE_SIZE )                   --Initial collisions are all zero.

--Initialise counters
word_total = 0
unique_total = 0
totals_by_length = repeat(0, MAX_WORD_SIZE)

--Hash mixer: Randomizes characters using a random translation. It reduces further the chance of hash collisions.
--Power lookup table: The hash mixer is extended to 2 dimensions, precalculating the power(2,i) multiplication.
--It also turns the one-time cypher into a one-time pad, as the rand() results are not reused.
sequence hash_mixer hash_mixer = repeat(repeat(0,MAX_CHAR), MAX_WORD_SIZE)
for i = 1 to MAX_WORD_SIZE do
    for c = '-' to 'Z' do --Of the possible token characters, '-' is the lowest used, and 'Z' the highest
        hash_mixer[i][c] = rand(1111)*power(2,i)
    end for
end for

-------------------------------------------------------------------------------
--Aggregation:
--Calculate the total number of words and unique words
--Figure out how many times each word is used
--Count the number of words for each length

--Go through each word
word = get_next_word()
while fn do
    --Get the word length
    word_length = length(word)
    --Compute a hash value, using the hash_mixer.
    hash = 0
    for i = 1 to word_length do
        hash = xor_bits(hash, hash_mixer[i][word[i]] )
    end for
    --remember the raw hash, makes it faster when comparing words
    raw_hash = hash
    --wrap the hash around so that it fits inside the table
    hash = remainder(hash, TABLE_SIZE ) + 1 
    
    --Check to see if it exists in the hash bucket already
    is_unique = 1
    --For each element in the bucket (Could possibly speed up with sorting, and a binary search method?)
    for i = 1 to table_width[hash] do   
        --If it exists (compare raw hashes first, if the same check the whole word)
        if hash_table[hash][i][RAW] = raw_hash and not compare(word, hash_table[hash][i][WORD]) then
            is_unique = 0               --Then it isn't unique
            index_in_bucket = i         --remember where it was in the bucket
            exit
        end if
    end for
    --If it doesn't exist yet, add it
    if is_unique then
        table_width[hash] += 1                      --increase the collisions count, and
        --store the new element in the table
        hash_table[hash][table_width[hash]] = { -1, raw_hash, word } 
        unique_total += 1                           --increment the number of unique words found
    else
    --If it does... increase its frequency count (yes, it should be '-=')
        hash_table[hash][index_in_bucket][FREQ] -= 1        
    end if
    --Update counters
    word_total += 1
    totals_by_length[word_length] += 1      
        
    --Get the next word from the file
    word = get_next_word()
end while
--</AGGREGATOR>================================================================

--<RANKING>====================================================================
--The hash table now contains all the words and their frequency counts
--Go through the hash table and find the most popular words...
-------------------------------------------------------------------------------
--'Constants'
--The number of words to show in the output
global integer RANK_SIZE
RANK_SIZE = floor(unique_total/1000)+1
if RANK_SIZE > 20 then RANK_SIZE = 20 
elsif not unique_total then RANK_SIZE = 0 
end if

--Find the largest word that was in the text
global integer LONGEST_WORD LONGEST_WORD = 0 
for i = MAX_WORD_SIZE to 1 by -1 do
    if totals_by_length[i] then
        LONGEST_WORD = i
        exit
    end if
end for

-------------------------------------------------------------------------------
--Array of the top RANK_SIZE records, from high(1) to low(20)
global sequence top_words   
-------------------------------------------------------------------------------
--Initialise arrays
top_words = repeat({0,0,{}}, RANK_SIZE)
-------------------------------------------------------------------------------
--Find the most commonly occuring words.
--top_words is a sorted list, where the most common word is the first in the list.
--The frequency is stored with the word itself
for bucket = 1 to TABLE_SIZE do
    for w = 1 to table_width[bucket] do --(skip if no elements in bucket)
        --For every record in the hash table...
        word = hash_table[bucket][w]

        --If the new word occurs more often than the lowest scoring word on the board...
        --(higher frequency, or equal freq's and alphabetically lower name)
        --(freq stored as a negative - means that compare will sort correctly)
        if compare(word, top_words[RANK_SIZE]) = -1 then
            --Then bump it off and replace it with the new word
            top_words[RANK_SIZE] = word

            --That new word is on the lowest rung now, but may belong higher.
            --Propagate upwards like a bubble sort until it has been fixed.
            for i = RANK_SIZE to 2 by -1 do
                --If a record scores higher than the one above it...
                if compare(top_words[i], top_words[i-1]) = -1 then
                    --Swap them!
                    word = top_words[i]
                    top_words[i] = top_words[i-1]
                    top_words[i-1] = word
                --Otherwise, the array is fully sorted (only 1 disturbance per cycle occurs)
                else
                    exit
                end if
            end for
        end if
    end for
end for
--</RANKING>====================================================================